/* File: bisect.c 
   Find a root for func() using function bisect() with bisection method 
   in the range [x0, xf]. The signs of f(x0) and f(xf) must be diff 
   to guarantee there is a root */
#include <stdio.h>
#include <math.h>              /* for fabs() */
#include <float.h>             /* for FLT_EPSILON */

/* function to find a root using bisection method */
int bisect(double *x, double x0, double xf, double tolerance, 
           double (*func)(double)) {
    double xleft = x0, fleft;  /* x and f(x) on the left of the root */
    double xright = xf;        /* x on the right of the root */
    double xmid, fmid;         /* x and f(x) on the mid of xleft and xright */

    if(tolerance <= 0.0)       /* invalid tolerance */
       return -1;              /* return -1 for invalid tolerance */

    fleft = func(x0);               /* calculate f(xleft) */
    if(fleft * func(xf) > 0.0) {    /* invalid x0 and xf */
       return -2;                   /* return -2 for invalid x0 and xf */
    }

    xmid = (xleft+xright)/2.0;      /* mid point */
    fmid = func(xmid);              /* f() in the mid point */
    while(fabs(fmid) > tolerance) {/* convergence criterion to terminate loop */
        /* f(xleft)*f(mid) to determine the updating of left or right point */
        if(fleft*fmid <= 0.0) {     /* mid point becomes the right point */
            xright = xmid;          /* update the right point */
        }
        else {                      /* mid point becomes the left point */
            xleft = xmid;           /* update the left point */
            fleft = fmid;           /* update f(xleft) */
        }
        xmid = (xleft+xright)/2.0;  /* update the mid point */
        fmid = func(xmid);          /* f() in the mid point */
    } 
    *x = xmid;                      /* pass xmid to the calling function */
    return 0;                       /* return 0 for successful */
}

int main() {
    /* find the root at M_PI for sin(x) specified in the range [2.0, 4.0].
       use machine epsilon for float type as convergence tolerance */
    double x, x0=2.0, xf=4.0, tolerance=FLT_EPSILON;
    int status;

    status = bisect(&x, x0, xf, tolerance, sin);
    if(status == -1)  /* calling bisect() failed */
        printf("Error: bisect() failed, invalid tolerance\n");
    else if(status == -2)  /* calling bisect() failed */
        printf("Error: bisect() failed, invalid end points\n");
    else            /* calling bisect() successfully */
        printf("x = %f, sin(x) = %g\n", x, sin(x));
    return 0;
}
