/************************************************************
* File: studentfunc.c
* This file contains function definitions for the GPA library
************************************************************/
#include <stdio.h>
#include <string.h>   /* for strcmp() */
#include "student.h"  /* header file GPA lib */

/* static function in file scope */
static
student_t *highestLowestGPA(const student_t *head, int isFindHighest);

/* external functions */

/* Code block 1: functions for a single student */
void gpa_printStudent(const student_t *sp) { /* print info for a student */
    if(sp == NULL) {
        printf("Error: invalid argument passed to printStudent()\n");
        return;
    }
    printf("  id: %d; name: %6s; birthday: %8s %2d, %d; GPA: %.2f\n",
           sp->id, sp->name, gpa_monthString(sp->birthday.month),
           sp->birthday.day, sp->birthday.year, sp->gpa);
}      

int gpa_getId(const student_t *sp) {      /* get student ID */
    return sp->id;
}

char *gpa_getName(const student_t *sp) {  /* get student name */
    return (char *)sp->name;
}

struct Birthday gpa_getBirthday(const student_t *sp) { /* get birthday */
    return sp->birthday;
}

double gpa_getGPA(const student_t *sp) {  /* get student GPA */
    return sp->gpa;
}


/* Code block 2: functions for a list of students */
/* get total number of students in the list */
int gpa_totnum(const student_t *head) { 
    int totnum = 0;
    while(head->id != GPA_EOD) {
        totnum++;
        head++;
    }
    return totnum;
}   

/* print info for all students in the list */
void gpa_printList(const student_t *head) {
    while(head->id != GPA_EOD) {
        gpa_printStudent(head);
        head++;
    }
}   

/* calculate the average of the GPA for students in the list */
double gpa_averageGPA(const student_t *head) {
    double gpa, totalval;
    int i;
 
    totalval=0.0;
    for(i=0; head->id != GPA_EOD; head++, i++) { 
         totalval += head->gpa;
    }
    gpa = totalval/i;
    return gpa;
}

/* find the student with the highest GPA in the list */
student_t *gpa_highestGPA(const student_t *head) {
    return highestLowestGPA(head, TRUE);
}
  
/* find the student with the lowest GPA in the list */
student_t *gpa_lowestGPA(const student_t *head) {
    return highestLowestGPA(head, FALSE);
}

/* find the student based on the id and name */
student_t *gpa_find(const student_t *head, int id, const char *name) {
    while(head->id!=GPA_EOD && !(!strcmp(head->name, name) && head->id == id)) {
       head++;
    }
    if(head->id == GPA_EOD)
       return NULL;
    else
       return (student_t *)head;
}

/* print out the students with GAP above 'gpa' */
int gpa_aboveGPA(const student_t *head, double gpa) {
    int num;
    for(num=0; head->id != GPA_EOD; head++) { 
        if(head->gpa >= gpa) {
            gpa_printStudent(head);
            num++;
        }
    }
    return num;
}

/* sort students with GPA in descending order using selection sort */
void gpa_sortGPA(student_t *head) {
    int i, j, totnum; /* declare indexes i and j, total num of students */
    int imax;         /* declare index imax for the maximum GPA */
    student_t tmp;    /* declare a temporary student element for swap */

    totnum = gpa_totnum(head); /* get the total number of students to sort */
    /* find the student with highest GPA in the remaining part of list
       in each iteration staring with head[i]  */
    for(i = 0; i < totnum-1; i++) {
        imax = i;
        /* compare with head[i].gpa with head[j].gpa for j from i+1
           to totnum-1 */
        for(j = i+1; j < totnum; j++) {
            if(head[imax].gpa < head[j].gpa) { 
                imax = j; /* save the index of the new maximum */
            }
        }
        if(imax != i) {   /* swap head[i] and head[imax] */
            tmp = head[i];
            head[i] = head[imax];
            head[imax] = tmp;
        }
    }
}


/* Code block 3: utility functions */

/* The function returns a string of a month based on the 
   argument in the range of 1 to 12. If the value of the 
   argument is outside this range, the function returns NULL. */
const char *gpa_monthString(int month) {
    static const char *monthName[] = {"January", "February", "March", 
           "April", "May", "June", "July", "August", "September", 
           "October", "November", "December"};

    return (month < 1 || month > 12) ? NULL : monthName[month-1];
}


/* Code block 4: internal functions */

/* find the student with the highest or lowest GPA in the list */
static
student_t *highestLowestGPA(const student_t *head, int isFindHighest) {
    student_t *sp;
    double gpa;

    gpa = head->gpa;       /* initial gpa */
    sp= (student_t *)head; /* initial sp */
    while(head->id != GPA_EOD) {  /* search through the list */
        /* if isFindHighest is TRUE, find the student with the 
           highest GPA, else find the student with the lowest GPA */
        if(isFindHighest == TRUE && head->gpa > gpa) {
            gpa = head->gpa;
            sp= (student_t *)head;
        }
        else if(isFindHighest == FALSE && head->gpa < gpa) {
            gpa = head->gpa;
            sp = (student_t*)head;
        }
        head++;  /* move the next element */
    }

    return sp;
}
