/*********************************************************************
* File: studentlistmenu.c
* This application program uses the GPA library with a linked list
* to process GPA's of students in the database of a file. 
* The application is menu driven.
**********************************************************************/
#include <stdio.h>
#include <string.h>             /* for strcmp()  */
#include <stdlib.h>             /* for exit(EXIT_FAILURE) */
#include "studentfile.h"        /* the header file for GPA lib */


#define MAX_TRY_NUM  3          /* maximum number of tries */
#define MAX_PROMPT_STR  128     /* maximum prompt string length */
#define MAX_FILENAME_LENGTH 128 /* maximum file name length */

/* enumerated type for actions from the menu driven user input */
typedef enum Action {
    UNDEFINED,   /* undefined action */
    AVERAGEGPA,  /* find the average of GPA's for all students */
    HIGHESTGPA,  /* find the student with the highest GPA */
    LOWESTGPA,   /* find the student with the lowest GPA */
    ABOVEGPA,    /* find students with GPA above a given value */
    SORTGPA,     /* sort the students based on GPA in descending order */
    OBTAINGPA,   /* obtain the GPA for a student based on id and name */
    CHANGEGPA,   /* change the GPA for a student based on id and name */ 
    READFILE,    /* read student information from a file */
    WRITEFILE,   /* write student information into a file */
    ADDSTUDENT,  /* add a student into the list */
    REMOVESTUDENT, /* remove a student from the list */
    QUIT         /* quit the program */
} action_t;

/* function prototypes for this application only */
extern action_t getActionMenu(void);  /* action menu */
/* obtain the GPA for a student */
extern void obtainGPA(StudentNode head); 
/* change the GPA for a student */
extern void changeGPA(StudentNode head); 
/* initialize the head of the list for students */
extern StudentNode initialize(const char *filename);  
/* add a student to the list */
extern void addStudent(StudentNode *head);
/* remove a student from the list */
extern void removeStudent(StudentNode *head);


int main() {
    StudentNode head, sn; /* head of the list and a single student */
    double gpa;           /* GPA value */
    struct Birthday b;    /* b for birthday */
    int num; /* number of students with GPA above a specified value */
    action_t act;         /* action in the menu */
    int quit = FALSE;     /* flag for quit or continue the menu */
    /* declare filename for data file, initialized with the default data file */
    char filename[MAX_FILENAME_LENGTH] = "student.data";

    /* head points to the list of students in the default data file student.data */
    head = initialize(filename);
    while(quit==FALSE) {  /* continue the menu if not quit */
        act = getActionMenu();  /* get an action */
        switch(act) {     /* take an action based on the value of 'act' */
            case AVERAGEGPA:  /* obtain the average of all GPA's */
              printf("The GPA for all students is %.2f\n", gpa_averageGPA(head));
              break;
            case HIGHESTGPA: /* find the student with the highest GPA */ 
              sn = gpa_highestGPA(head);
              b = gpa_getBirthday(sn);
              printf("%s who was born on %s %d, %d has the highest GPA %.2f\n",
                     gpa_getName(sn), gpa_monthString(b.month), b.day, b.year,
                     gpa_getGPA(sn));
              break;
            case LOWESTGPA:  /* find the student with the lowest GPA */
              sn = gpa_lowestGPA(head);
              b = gpa_getBirthday(sn);
              printf("%s who was born on %s %d, %d has the lowest GPA %.2f\n",
                     gpa_getName(sn), gpa_monthString(b.month), b.day, b.year,
                     gpa_getGPA(sn));
              break;
            case ABOVEGPA: /* find students with GPA above a given value */
              printf("please input GPA\n");
              scanf("%lf%*c", &gpa);
              printf("Students with GPA %.2f or above:\n", gpa);
              num = gpa_aboveGPA(head, gpa);
              printf("%d students with GPA %.2f or above.\n", num, gpa);
              break;
            /* sort the students based on GPA in descending order */
            case SORTGPA:
              gpa_sortGPA(head);  /* sort in descending order */
              printf("The student record sorted in descending GPA:\n");
              gpa_printList(head);/* print all students */
              break;
            /* obtain the GPA for a student based on id and name */
            case OBTAINGPA:
              obtainGPA(head);    /* using a function in this program */
              break;
            /* change the GPA for a student based on id and name */
            case CHANGEGPA:
              changeGPA(head);
              break;
            /* read the student information in a data file */
            case READFILE:
              printf("Please input a GPA data file name\n");
              fgets(filename, MAX_FILENAME_LENGTH, stdin); /* get file name */
              filename[strlen(filename)-1] = '\0';         /* remove '\n' */
              gpa_clearList(head);  /* release the allocated memory for head */
              /* head points to the list of students from filename */
              head = initialize(filename);  
              break;
            /* write the student information in the list into a data file */
            case WRITEFILE:
              printf("Please input a GPA data file name\n");
              fgets(filename, MAX_FILENAME_LENGTH, stdin); /* get file name */
              filename[strlen(filename)-1] = '\0';         /* remove '\n' */
              gpa_writeData(head, filename);  /* write list into filename */
              break;
            /* add a student to the list */
            case ADDSTUDENT:
              addStudent(&head);
              break;
            /* remove a student from the list */
            case REMOVESTUDENT:
              removeStudent(&head);
              break;
            case QUIT:            /* quit the program */
              quit = TRUE;        /* set flag quit to TRUE */
              break;
            default:              /* in case a mistake is made */
              printf("\aInvalid input!\n");
              break;
        }
    }
    gpa_clearList(head);  /* release the allocated memory for head */
    return 0;
}

/*-----------------------------------------------------------------
* Function getActionMenu() gets the action item from user' menu.
* The function returns a proper enum in enum Action, if successful.
*                      UNDEFINED                   , if failed.
*-----------------------------------------------------------------*/
action_t getActionMenu(void) {
    action_t act;                  /* action to be taken */
    const char *prompt = "GPA> ";  /* prompt for the menu */
    char str[MAX_PROMPT_STR];      /* place holder for the user input */

    printf("\n");
    printf("************************ GPA Menu ************************\n");
    printf("av, average: display the average of GPAs for all students\n");
    printf("h,  highest: display the student with the highest GPA\n");
    printf("l,  lowest:  display the student with the lowest GPA\n");
    printf("ab, above:   display students with GPA above a value\n");
    printf("s,  sort:    display the record sorted in descending GPA\n");
    printf("o,  obtain:  obtain a student's GPA\n");
    printf("c,  change:  change a student's GPA\n");
    printf("r,  read:    read a new GPA data file\n");
    printf("w,  write:   write to a GPA data file\n");
    printf("a,  add:     add a student\n");
    printf("re, remove:  remove a student\n");
    printf("q,  quit:    quit the system\n");
    printf("..........................................................\n\n");

    printf("%s", prompt);             /* print the prompt "GPA> " */
    fgets(str,MAX_PROMPT_STR,stdin);  /* get the user input */
    str[strlen(str)-1] = '\0';  /* get rid of '\n' in the input string */
    /* set act based on the user input */
    /* if the user input 'av' or 'average' */
    if(!strcmp(str, "av") || !strcmp(str, "average")) {
        act = AVERAGEGPA;
    }
    /* else if the user input 'h' or 'highest' */
    else if(!strcmp(str, "h") || !strcmp(str, "highest")) {
        act = HIGHESTGPA;
    }
    else if(!strcmp(str, "l") || !strcmp(str, "lowest")) {
        act = LOWESTGPA;
    }
    else if(!strcmp(str, "ab") || !strcmp(str, "above")) {
        act = ABOVEGPA;
    }
    else if(!strcmp(str, "s") || !strcmp(str, "sort")) {
        act = SORTGPA;
    }
    else if(!strcmp(str, "o") || !strcmp(str, "obtain")) {
        act = OBTAINGPA;
    }
    else if(!strcmp(str, "c") || !strcmp(str, "change")) {
        act = CHANGEGPA;
    }
    else if(!strcmp(str, "r") || !strcmp(str, "read")) {
        act = READFILE;
    }
    else if(!strcmp(str, "w") || !strcmp(str, "write")) {
        act = WRITEFILE;
    }
    else if(!strcmp(str, "a") || !strcmp(str, "add")) {
        act = ADDSTUDENT;
    }
    else if(!strcmp(str, "re") || !strcmp(str, "remove")) {
        act = REMOVESTUDENT;
    }
    else if(!strcmp(str, "q") || !strcmp(str, "quit")) {
        act = QUIT;
    }
    else { /* none of the above valid input */
        act = UNDEFINED;
    }
    return act;
}


/*--------------------------------------------------------------------
* Function obtainGPA() Obtains the GPA for a student from the database.
* Make sure the student with id and name is in the database.
* The user can input id and name up to 3 times defined in
* MAX_TRY_NUM. After three failed tries, the function will quit.
* The static variable try_count is used for implementation.
*---------------------------------------------------------------------*/
void obtainGPA(StudentNode head) {
    StudentNode sn;        /* a student */
    int id;                /* student id */
    char name[32];         /* student name */
    static try_count = 0;  /* track the number of tries using static var */

    printf("please input student id\n");
    /* get the student id and suppress the newline character '\n' */
    scanf("%d%*c", &id);   
    printf("please input student name\n");
    fgets(name, sizeof(name), stdin); /* the user input string */
    /* get rid of the newline character '\n' in the input string */
    name[strlen(name)-1] = '\0';      
    sn = gpa_find(head, id, name); /* find the student based on id and name */
    if(sn) { /* when the student is found, print his GPA */
        printf("%s's GPA is %.2f\n", name, gpa_getGPA(sn));
        try_count = 0;
    }
    else {  /* when the student is not found, try it again */
        printf("%s with id %d is not in the database.\n", name, id);
        try_count++; /* increment the count for number of failed try */
        if(try_count < MAX_TRY_NUM) { /* if less than specified num of tries */
           obtainGPA(head); /* call the function recursively, up to 3 times */
        }
        else {              /* exit the function after three failed tries */ 
           printf("You have tried %d times, nice try!\n", try_count);
           try_count = 0;   /* reset the count for next action */
        }
    }
}

/*-------------------------------------------------------------
* Function initialize() creates a list for student information
* in file filename. The memory allocated for the list needs to
* be released by function gpa_clearList().
*------------------------------------------------------------*/
StudentNode initialize(const char *filename) {
    StudentNode head;
    head = gpa_readData(filename);
    if(head == NULL) {
        fprintf(stderr, "Error: failed to read file '%s'\n", filename);
    }
    return head;
}

/*---------------------------------------------------------------------
* Function changeGPA() changes the GPA for a student from the database
* of a file. Make sure the student with id and name is in the database.
* The user can input id and name up to 3 times defined in * MAX_TRY_NUM.
* After three failed tries, the function will quit.
* The static variable try_count is used for implementation.
*---------------------------------------------------------------------*/
void changeGPA(StudentNode head) {
    StudentNode sn;         /* a student */
    int id;                /* student id */
    char name[32];         /* student name */
    static try_count = 0;  /* track the number of tries using static var */
    double gpa;            /* new GPA value from the user input */

    printf("please input student id\n");
    /* get the student id and suppress the newline character '\n' */
    scanf("%d%*c", &id);   
    printf("please input student name\n");
    fgets(name, sizeof(name), stdin); /* the user input string */
    /* get rid of the newline character '\n' in the input string */
    name[strlen(name)-1] = '\0';      
    sn = gpa_find(head, id, name); /* find the student based on id and name */
    if(sn) { /* when the student is found, change the GPA */
        printf("please input GPA\n");
        scanf("%lf%*c", &gpa);     /* get GPA and suppress RETURN character */
        gpa_setGPA(sn, gpa);  /* change the GPA in the list for the student */
        printf("%s's GPA has been changed to %.2f\n", name, gpa);
        try_count = 0;
    }
    else {  /* when the student is not found, try it again */
        printf("%s with id %d is not in the database.\n", name, id);
        try_count++; /* increment the count for number of failed try */
        if(try_count < MAX_TRY_NUM) { /* if less than specified num of tries */
           changeGPA(head); /* call the function recursively, up to 3 times */
        }
        else {              /* exit the function after three failed tries */ 
           printf("You have tried %d times, nice try!\n", try_count);
           try_count = 0;   /* reset the count for next action */
        }
    }
}

/*------------------------------------------------------------------
* Function addStudent() adds a student to the linked list inside the 
* system. Make sure the student with id and name is not in the 
* database. The user can input id and name up to 3 times defined in
* MAX_TRY_NUM. After three failed tries, the function will quit.
* The static variable try_count is used for implementation. 
*-------------------------------------------------------------------*/
void addStudent(StudentNode *head) {
    StudentNode sn;        /* a student */
    int id;                /* student id */
    char name[32];         /* student name */
    static try_count = 0;  /* track the number of tries using static var */
    double gpa;            /* new GPA value from the user input */
    struct Birthday birthday; /* structure for birthday info */

    sn = *head;            /* get the head of the list */
    printf("please input student id\n");
    /* get the student id and suppress the newline character '\n' */
    scanf("%d%*c", &id);
    printf("please input student name\n");
    fgets(name, sizeof(name), stdin); /* the user input string */
    /* get rid of the newline character '\n' in the input string */
    name[strlen(name)-1] = '\0';
    sn = gpa_find(*head, id, name); /* find the student based on id and name */
    if(sn) { /* id is already in the database, try MAX_TRY_NUM times */
        printf("Id %d is in the database already.\n", id);
        try_count++; /* increment the count for number of failed try */
        if(try_count < MAX_TRY_NUM) { /* if less than specified num of tries */
           addStudent(head);/* call the function recursively, up to 3 times */
        }
        else {              /* exit the function after three failed tries */
           printf("You have tried %d times, nice try!\n", try_count);
           try_count = 0;   /* reset the count for next action */
        }
    }
    else {
        printf("please input student birthday (day month year) \n");
        scanf("%hd%hd%hd%*c", &birthday.day, &birthday.month, &birthday.year);
        printf("please input GPA\n");
        scanf("%lf%*c", &gpa); /* get the GPA */
        /* create a student node */
        sn = gpa_createStudent(id, name, birthday, gpa); 
        gpa_append(head, sn); /* append the new student to the list */
        try_count = 0;        /* reset the count for next action */
    }
}

/*----------------------------------------------------------------------
* Function removeStudent() removes a student from the linked list inside 
* the system.  Make sure the student with id and name is in the database
* The user can input id and name up to 3 times defined in
* MAX_TRY_NUM. After three failed tries, the function will quit.
* The static variable try_count is used for implementation. 
*----------------------------------------------------------------------*/
void removeStudent(StudentNode *head) {
    StudentNode sn;        /* a student */
    int id;                /* student id */
    char name[32];         /* student name */
    static try_count = 0;  /* track the number of tries using static var */

    printf("please input student id\n");
    /* get the student id and suppress the newline character '\n' */
    scanf("%d%*c", &id);
    printf("please input student name\n");
    fgets(name, sizeof(name), stdin); /* the user input string */
    /* get rid of the newline character '\n' in the input string */
    name[strlen(name)-1] = '\0';
    sn = gpa_find(*head, id, name); /* find the student based on id and name */
    if(sn) { /* when the student is found, remove the student data from the list */
        gpa_remove(head, sn);   /* remove the student */
        try_count = 0;   /* reset the count for next action */
    }
    else {  /* when the student is not found, try it again */
        printf("%s with id %d is not in the database.\n", name, id);
        try_count++; /* increment the count for number of failed try */
        if(try_count < MAX_TRY_NUM) { /* if less than specified num of tries */
           removeStudent(head); /* call the function recursively, up to 3 times */
        }
        else {              /* exit the function after three failed tries */
           printf("You have tried %d times, nice try!\n", try_count);
           try_count = 0;   /* reset the count for next action */
        }
    }
}
